package fr.doap.jdb;

import fr.doap.jdb.impl.*;
import fr.doap.jdb.utils.*;
import fr.doap.slog.*;

import java.lang.reflect.*;
import java.sql.*;
import java.util.*;

import static fr.doap.jdb.utils.Utils.*;

public class Objects2Maps extends ObjectVisitor {
	
	public Objects2Maps(DBWOC aWOC) { mDBWOC =aWOC; }
	
	public Map<Object, Map<String, Object>> getInserts() { return mInserts; }
	public Map<Object, Map<String, Object>> getUpdates() { return mUpdates; }
	public Map<String, String> getColumnSQLTypes() { return mColumnSQLType; }
	
	@Override
	protected void onObjectAdded(Object aObject, int aPos) {
		if (mDBWOC.getId(aObject) == -1) {
			mDBWOC.addNew(aObject);
			mNewObjects.add(aObject);
		}
	}
	
	@Override
	protected void onProcessingObject(Object aObject, int aPos) {
		Class<?> lClass = aObject.getClass();
		try {
			Constructor lCons = lClass.getDeclaredConstructor(); // ensure at least a default constructor...
		} catch (Exception aE) {
			throw new RuntimeException("No constructor for class "+lClass.getCanonicalName());
		}
		
		Set<Object> lReachableObjects = new HashSet<>();
		mReach.put(aObject, lReachableObjects);
		
		Map<String, Object> lProps = getProperties(aObject);
		
		for (Map.Entry<String, Object> iME: lProps.entrySet()) {
			Object lV = iME.getValue();
			if (!sTypeMap.containsKey(lV.getClass())) {
				add(lV); // process object if necessary
				iME.setValue(mDBWOC.getId(lV)); // replace object by ID
				mColumnSQLType.put(iME.getKey(), "INTEGER"); // column type
				lReachableObjects.add(lV); // add object to reachable list
			} else mColumnSQLType.put(iME.getKey(), sTypeMap.get(lV.getClass()));
		}
		
		lProps.put(DBWrap.sJID, mDBWOC.getId(aObject));
		lProps.put("_CLASS_", lClass.getCanonicalName());
		
		if (mNewObjects.contains(aObject)) mInserts.put(aObject, lProps);
		else mUpdates.put(aObject, lProps);
	}
	
	@Override
	protected void onEndingRun(int aStart, int aPos) {
		// saturate
		Set<Object> lReachChanged = new HashSet<>(mReach.keySet());
		while (lReachChanged.size() > 0) {
			Set<Object> lNewReachChanged = new HashSet<>();
			for (Object iKey : lReachChanged) {
				Set<Object> lSet = mReach.get(iKey);
				boolean lSetUpdated = false;
				Set<Object> lNew = new HashSet<>(lSet);
				for (Object iO : lSet) if (lNew.addAll(mReach.get(iO))) lSetUpdated = true;
				if (lSetUpdated) {
					lSet.clear();
					lSet.addAll(lNew);
					lNewReachChanged.add(iKey);
				}
			}
			lReachChanged = lNewReachChanged;
		}

		mReach.forEach((aO, aSet) -> {
			if (aSet.size() > 0) {
				Object[] lObs = aSet.toArray();
				int lNew[] = new int[lObs.length];
				for (int i = 0; i < lNew.length; i++) lNew[i] = mDBWOC.getId(lObs[i]);
				L.i("Reach {} -> {}", aSet, lNew);
				if (mInserts.containsKey(aO)) mInserts.get(aO).put("_SUBS_", Utils.toByteArray(lNew));
				else mUpdates.get(aO).put("_SUBS_", Utils.toByteArray(lNew));
			}
		});
		
		L.i("Reach: {}", mReach);
	}
	
	protected Map<String, Object> getProperties(Object aObject) {
		Map<String, Object> lProps = new HashMap<>();
		Class<?> lC = aObject.getClass();
		while (lC != Object.class) {
			lProps.put(getShortClassname(lC), true);
			if (aObject instanceof Collection && lC.getCanonicalName().startsWith("java.")) {
				Collection lColl = (Collection) aObject;
				Object[] lObjects = new Object[2*lColl.size()];
				Iterator iIt = lColl.iterator();
				for (int i=0; i<lColl.size(); i++) {
					Object lV = iIt.next();
					if (!sTypeMap.containsKey(lV.getClass())) {
						add(lV);
						lObjects[2*i] = null;
						lObjects[2*i+1] = mDBWOC.getId(lV);
					} else {
						lObjects[2*i] = lV.getClass().getCanonicalName();
						lObjects[2*i+1] = lV;
					}
				}
				L.i("Coded collection is {}", Arrays.asList(lObjects));
				lProps.put("_CONTENT_", Utils.toByteArray(lObjects));
				break;
			} else {
				Field[] lFields = lC.getDeclaredFields();
				for (Field iF : lFields) {
					int lModif = iF.getModifiers();
					if (((lModif & Modifier.FINAL) | (lModif & Modifier.TRANSIENT) | (lModif & Modifier.VOLATILE)) == 0) {
						if (iF.trySetAccessible()) {
							try {
								lProps.put(getShortClassname(lC) + "#" + iF.getName(), iF.get(aObject));
							} catch (Exception aE) {
								L.t(aE);
							}
						}
					}
				}
				lC = lC.getSuperclass();
			}
		}
		return lProps;
	}
	
	DBWOC 								mDBWOC;

	Set<Object>							mNewObjects = new HashSet<>();
	Map<Object, Set<Object>>			mReach =new HashMap<>();
	Map<Object, Map<String, Object>>	mInserts =new HashMap<>();
	Map<Object, Map<String, Object>>	mUpdates =new HashMap<>();
	Map<String, String>					mColumnSQLType =new HashMap<>();
	
	static final Map<Class<?>, String> sTypeMap = Utils.HashMapOf(
		Boolean.class,	"BOOLEAN",
		Integer.class,	"INTEGER",
		Long.class,		"BIGINT",
		byte[].class,	"VARBINARY",
		Blob.class,		"BLOB",
		String.class,	"VARCHAR" );
}
