package fr.doap.jdb.impl;

import fr.doap.slog.*;

import java.lang.reflect.*;
import java.util.*;

abstract
public class ClassAnalyzer {
	
	public boolean isApplicable(Class<?> aClass) { return true; }
	public boolean processSuperclass(Class<?> aClass) { return true; }
	
	abstract public void analyze(Class<?> aClass, Map<String, PropGetterSetter> aMap);

	
	static public class FieldCA extends ClassAnalyzer {
		@Override
		public void analyze(Class<?> aClass, Map<String, PropGetterSetter> aMap) {
			try {
				Field[] lFields = aClass.getDeclaredFields();
				for (Field iF: lFields) {
					if (iF.trySetAccessible()) {
						PropGetterSetter lPGS = new PropGetterSetter.FieldGS(iF);
						if (lPGS != null) aMap.put(iF.getName(), lPGS);
					}
				}
			} catch (Exception aE) { L.t(aE); }
		}
	}
	
	static public class CollectionCA extends ClassAnalyzer {
		@Override
		public boolean isApplicable(Class<?> aClass) {
			return (Collection.class.isAssignableFrom(aClass) && aClass.getCanonicalName().startsWith("java."));
		}
		
		@Override
		public boolean processSuperclass(Class<?> aClass) { return false; }
		
		@Override
		public void analyze(Class<?> aClass, Map<String, PropGetterSetter> aMap) {
			aMap.put("_CONTENT_", new PropGetterSetter.CollectionGS());
		}
	}
}
